/** Copyright (C) 2017 by Bundesanstalt für Gewässerkunde
 * Software engineering by
 *  Björnsen Beratende Ingenieure GmbH
 *  Dr. Schumacher Ingenieurbüro für Wasser und Umwelt
 *
 * This file is Free Software under the GNU AGPL (>=v3)
 * and comes with ABSOLUTELY NO WARRANTY! Check out the
 * documentation coming with Dive4Elements River for details.
 */

package org.dive4elements.river.artifacts.uinfo.salix;

import java.util.HashSet;
import java.util.Set;

import org.dive4elements.artifactdatabase.state.ArtifactAndFacet;
import org.dive4elements.artifactdatabase.state.Facet;
import org.dive4elements.artifacts.CallContext;
import org.dive4elements.river.artifacts.D4EArtifact;
import org.dive4elements.river.artifacts.access.RangeAccess;
import org.dive4elements.river.artifacts.common.AbstractProcessor;
import org.dive4elements.river.artifacts.model.CrossSectionWaterLineFacet;
import org.dive4elements.river.artifacts.model.FacetTypes;
import org.dive4elements.river.artifacts.resources.Resources;
import org.dive4elements.river.artifacts.states.DefaultState.ComputeType;
import org.dive4elements.river.artifacts.uinfo.common.UInfoResultType;
import org.dive4elements.river.artifacts.uinfo.salix.SalixLineCrossSectionIndexData.SalixWaterlevel;
import org.dive4elements.river.exports.DiagramGenerator;
import org.dive4elements.river.themes.ThemeDocument;

/**
 * Processor to generate the facets and data series of salix line
 *
 * @author Matthias Schäfer
 *
 */
public final class SalixLineProcessor extends AbstractProcessor {

    private static final String FACET_SALIX_LINE_FILTERED = "uinfo_facet_salix_line.filtered";

    private static final String FACET_SALIX_LINE_FILTERED_DESCRIPTION = "uinfo_facet_salix_line.filtered.description";

    public static final String FACET_SALIX_LINE_RAW = "uinfo_facet_salix_line";

    private static final String FACET_SALIX_LINE_RAW_DESCRIPTION = "uinfo_facet_salix_line.raw.description";

    private static final String FACET_SALIX_MNWMW_FILTERED = "uinfo_facet_salix_mnwmw.filtered";

    private static final String FACET_SALIX_MNWMW_FILTERED_DESCRIPTION = "uinfo_facet_salix_mnwmw.filtered.description";

    public static final String FACET_SALIX_MNWMW_RAW = "uinfo_facet_salix_mnwmw";

    private static final String FACET_SALIX_MNWMW_RAW_DESCRIPTION = "uinfo_facet_salix_mnwmw.raw.description";

    public static final String FACET_SALIX_SCENARIO_FILTERED = "uinfo_facet_salix_scenario.filtered";

    private static final String FACET_SALIX_SCENARIO_FILTERED_DESCRIPTION = "uinfo_facet_salix_scenario.filtered.description";

    public static final String FACET_SALIX_SCENARIO_RAW = "uinfo_facet_salix_scenario";

    private static final String FACET_SALIX_SCENARIO_RAW_DESCRIPTION = "uinfo_facet_salix_scenario.raw.description";

    public static final String FACET_SALIX_RANK = "uinfo_facet_salix_rank";

    private static final String FACET_SALIX_RANK_DESCRIPTION = "uinfo_facet_salix_rank.description";

    // REMARK: must end with FacetTypes.CROSS_SECTION_WATER_LINE so its correctly recognized as such a theme
    private static final String FACET_SALIX_CROSS_SECTION_WATER_LINE = "uinfo_facet_salix_line_" + FacetTypes.CROSS_SECTION_WATER_LINE;

    private static final String FACET_SALIX_CROSS_SECTION_WATER_LINE_DESCRIPTION = FACET_SALIX_CROSS_SECTION_WATER_LINE + "_description";

    private static final String FACET_SALIX_SCENARIO_CROSS_SECTION_WATER_LINE = "uinfo_facet_salix_scenario_" + FacetTypes.CROSS_SECTION_WATER_LINE;

    private static final String FACET_SALIX_WATERLEVEL_CROSS_SECTION_WATER_LINE_PREFIX = "uinfo_facet_salix_waterlevel_";

    private static final String I18N_AXIS_LABEL = "uinfo.chart.salix_line.section.yaxis.label";

    private static final Set<String> HANDLED_FACET_TYPES = new HashSet<>();

    static {
        HANDLED_FACET_TYPES.add(FACET_SALIX_LINE_FILTERED);
        HANDLED_FACET_TYPES.add(FACET_SALIX_LINE_RAW);
        HANDLED_FACET_TYPES.add(FACET_SALIX_MNWMW_FILTERED);
        HANDLED_FACET_TYPES.add(FACET_SALIX_MNWMW_RAW);
        HANDLED_FACET_TYPES.add(FACET_SALIX_SCENARIO_FILTERED);
        HANDLED_FACET_TYPES.add(FACET_SALIX_SCENARIO_RAW);
        HANDLED_FACET_TYPES.add(FACET_SALIX_RANK);
    }

    public SalixLineProcessor() {
        super(I18N_AXIS_LABEL, HANDLED_FACET_TYPES);
    }

    public static Facet createSalixLineFilteredFacet(final CallContext context, final String hash, final String id, final int facetIndex,
            final int resultIndex) {

        final String description = Resources.getMsg(context.getMeta(), FACET_SALIX_LINE_FILTERED_DESCRIPTION, FACET_SALIX_LINE_FILTERED_DESCRIPTION);
        return new SalixLineResultFacet(facetIndex, resultIndex, FACET_SALIX_LINE_FILTERED, description, I18N_AXIS_LABEL, id, hash);
    }

    public static Facet createSalixLineRawFacet(final CallContext context, final String hash, final String id, final int facetIndex, final int resultIndex) {

        final String description = Resources.getMsg(context.getMeta(), FACET_SALIX_LINE_RAW_DESCRIPTION, FACET_SALIX_LINE_RAW_DESCRIPTION);
        return new SalixLineResultFacet(facetIndex, resultIndex, FACET_SALIX_LINE_RAW, description, I18N_AXIS_LABEL, id, hash);
    }

    public static Facet createSalixMnwMwFilteredFacet(final CallContext context, final String hash, final String id, final int facetIndex,
            final int resultIndex) {

        final String description = Resources.getMsg(context.getMeta(), FACET_SALIX_MNWMW_FILTERED_DESCRIPTION, FACET_SALIX_MNWMW_FILTERED_DESCRIPTION);
        return new SalixMnwMwResultFacet(facetIndex, resultIndex, FACET_SALIX_MNWMW_FILTERED, description, I18N_AXIS_LABEL, id, hash);
    }

    public static Facet createSalixMnwMwRawFacet(final CallContext context, final String hash, final String id, final int facetIndex, final int resultIndex) {

        final String description = Resources.getMsg(context.getMeta(), FACET_SALIX_MNWMW_RAW_DESCRIPTION, FACET_SALIX_MNWMW_RAW_DESCRIPTION);
        return new SalixMnwMwResultFacet(facetIndex, resultIndex, FACET_SALIX_MNWMW_RAW, description, I18N_AXIS_LABEL, id, hash);
    }

    public static Facet createSalixScenarioFilteredFacet(final CallContext context, final String hash, final String id, final int dataIndex,
            final int resultIndex, final int facetIndex, final String subLabel) {

        final String description = Resources.getMsg(context.getMeta(), FACET_SALIX_SCENARIO_FILTERED_DESCRIPTION, FACET_SALIX_SCENARIO_FILTERED_DESCRIPTION,
                subLabel);
        return new SalixScenarioResultFacet(facetIndex, resultIndex, dataIndex, FACET_SALIX_SCENARIO_FILTERED, description, I18N_AXIS_LABEL, hash, id);
    }

    public static Facet createSalixScenarioRawFacet(final CallContext context, final String hash, final String id, final int dataIndex, final int resultIndex,
            final int facetIndex, final String subLabel) {

        final String description = Resources.getMsg(context.getMeta(), FACET_SALIX_SCENARIO_RAW_DESCRIPTION, FACET_SALIX_SCENARIO_RAW_DESCRIPTION, subLabel);
        return new SalixScenarioResultFacet(facetIndex, resultIndex, dataIndex, FACET_SALIX_SCENARIO_RAW, description, I18N_AXIS_LABEL, hash, id);
    }

    public static Facet createSalixRankFacet(final CallContext context, final String hash, final String id, final int facetIndex, final int resultIndex) {

        final String description = Resources.getMsg(context.getMeta(), FACET_SALIX_RANK_DESCRIPTION, FACET_SALIX_RANK_DESCRIPTION);
        return new SalixRankResultFacet(facetIndex, resultIndex, FACET_SALIX_RANK, description, I18N_AXIS_LABEL, id, hash);
    }

    public static Facet createSalixCrossSectionFacet(final CallContext context, final ComputeType type, final String hash, final String stateId,
            final int facetIndex) {
        final String description = Resources.getMsg(context.getMeta(), FACET_SALIX_CROSS_SECTION_WATER_LINE_DESCRIPTION);
        return new CrossSectionWaterLineFacet(facetIndex, FACET_SALIX_CROSS_SECTION_WATER_LINE, description, type, hash, stateId,
                SalixLineCrossSectionIndexData.salixLine());
    }

    public static Facet createSalixScenarioCrossSectionFacet(final CallContext context, final ComputeType type, final String hash, final String stateId,
            final int scenarioIndex, final int facetIndex, final String subLabel) {

        final String description = Resources.getMsg(context.getMeta(), FACET_SALIX_SCENARIO_FILTERED_DESCRIPTION, FACET_SALIX_SCENARIO_FILTERED_DESCRIPTION,
                subLabel);

        return new CrossSectionWaterLineFacet(facetIndex, FACET_SALIX_SCENARIO_CROSS_SECTION_WATER_LINE, description, type, hash, stateId,
                SalixLineCrossSectionIndexData.scenario(scenarioIndex));
    }

    public static Facet createSalixWaterlevelCrossSectionFacet(final CallContext context, final ComputeType type, final String hash, final String stateId,
            final SalixWaterlevel knownWaterlevel, final int facetIndex) {
        final String facetName = FACET_SALIX_WATERLEVEL_CROSS_SECTION_WATER_LINE_PREFIX + knownWaterlevel.name() + "_" + FacetTypes.CROSS_SECTION_WATER_LINE;
        final String facetI10n = facetName + "_description";
        final String facetDescirption = Resources.getMsg(context.getMeta(), facetI10n);

        return new CrossSectionWaterLineFacet(facetIndex, facetName, facetDescirption, type, hash, stateId,
                SalixLineCrossSectionIndexData.waterlevel(knownWaterlevel));
    }

    @Override
    protected String generateSeries(final DiagramGenerator generator, final ArtifactAndFacet bundle, final ThemeDocument theme, final boolean visible) {

        // log.debug("Processing facet: " + bundle.getFacetName());

        if (bundle.getFacetName().equals(FACET_SALIX_SCENARIO_FILTERED) || bundle.getFacetName().equals(FACET_SALIX_SCENARIO_RAW)) {
            final CallContext context = generator.getContext();
            final SalixLineCalculationResult data = (SalixLineCalculationResult) bundle.getData(context);
            final int dataIndex = ((SalixScenarioResultFacet) bundle.getFacet()).getDataIndex();
            final double[][] scenarioPoints = data.getScenarioPoints(dataIndex);
            return buildSeriesForPoints(scenarioPoints, generator, bundle, theme, visible, GAP_DISTANCE);
        }

        if (bundle.getFacetName().equals(FACET_SALIX_RANK)) {
            final RangeAccess kmRange = new RangeAccess((D4EArtifact) bundle.getArtifact());
            // TODO get all rank data (5 ranks) from database
            final double[][] points = new double[2][2];
            points[0][0] = kmRange.getLowerKm();
            points[0][1] = kmRange.getUpperKm();
            points[1][0] = 1.5;
            points[1][1] = 1.5;
            return buildSeriesForPoints(points, generator, bundle, theme, visible, null);
        }

        return buildSeriesForType(generator, bundle, theme, visible, doGetType(bundle.getFacetName()), GAP_DISTANCE);
    }

    protected UInfoResultType doGetType(final String facetName) {

        if (FACET_SALIX_LINE_FILTERED.contentEquals(facetName))
            return UInfoResultType.salixline;
        if (FACET_SALIX_LINE_RAW.contentEquals(facetName))
            return UInfoResultType.salixline;

        if (FACET_SALIX_MNWMW_FILTERED.contentEquals(facetName))
            return UInfoResultType.salix_mw_mnw;
        if (FACET_SALIX_MNWMW_RAW.contentEquals(facetName))
            return UInfoResultType.salix_mw_mnw;

        final String error = String.format("Unknown facet name: %s", facetName);
        throw new UnsupportedOperationException(error);
    }
}